import { __decorate } from "tslib";
import { nodeStrokeWidth } from './StatusIcons';
import { Result } from '../PipelineGraphModel';
import { LitElement, customElement, svg, } from 'lit-element';
import { styleMap } from 'lit-html/directives/style-map';
function polarToCartesian(centerX, centerY, radius, angleInDegrees) {
    const angleInRadians = ((angleInDegrees - 90) * Math.PI) / 180.0;
    return {
        x: centerX + radius * Math.cos(angleInRadians),
        y: centerY + radius * Math.sin(angleInRadians),
    };
}
function describeArcAsPath(x, y, radius, startAngle, endAngle) {
    const start = polarToCartesian(x, y, radius, endAngle);
    const end = polarToCartesian(x, y, radius, startAngle);
    const arcSweep = endAngle - startAngle <= 180 ? '0' : '1';
    const d = ['M', start.x, start.y, 'A', radius, radius, 0, arcSweep, 0, end.x, end.y].join(' ');
    return d;
}
let SvgSpinner = class SvgSpinner extends LitElement {
    constructor() {
        super(...arguments);
        this.percentage = 0;
        this.radius = 0;
        this.result = ``;
        // Data
        this.infiniteRotationRunning = false;
        this.infiniteRotateDegrees = 0;
        this.isEdgeOrIE = ('MSInputMethodContext' in window && 'documentMode' in document) || window.navigator.userAgent.indexOf('Edge') > -1;
        this.requestAnimationFrameId = 0; // Callback handle
        this.infiniteLoadingTimer = () => {
            this.infiniteRotateDegrees += 1.5;
            if (this.infiniteRotateDegrees >= 360) {
                this.infiniteRotateDegrees = 0;
            }
            this.animatedElement.setAttribute('transform', `rotate(${this.infiniteRotateDegrees})`);
            this.requestAnimationFrameId = requestAnimationFrame(this.infiniteLoadingTimer);
        };
    }
    // Properties
    static get properties() {
        return {
            radius: { type: Number, },
            result: { type: String, },
            percentage: { type: Number, },
        };
    }
    // Lifecycle
    createRenderRoot() {
        return this;
    }
    connectedCallback() {
        this.infiniteRotationRunning = false;
        this.infiniteRotateDegrees = 0;
        super.connectedCallback();
    }
    disconnectedCallback() {
        cancelAnimationFrame(this.requestAnimationFrameId);
        super.disconnectedCallback();
    }
    render() {
        const { result } = this;
        const radius = this.radius || 12;
        const insideRadius = radius - 0.5 * nodeStrokeWidth;
        let percentage = this.percentage;
        const groupClasses = ['PWGx-progress-spinner', result];
        if (result === Result.queued) {
            percentage = 0;
        }
        else if (typeof percentage !== 'number' || isNaN(percentage) || percentage < 0) {
            percentage = 0;
        }
        else if (percentage === 100) {
            groupClasses.push('pc-over-100');
            percentage = 0;
        }
        else if (percentage > 100) {
            groupClasses.push('spin');
            percentage = 25;
            if (!this.infiniteRotationRunning && this.isEdgeOrIE) {
                requestAnimationFrame(this.infiniteLoadingTimer);
                this.infiniteRotationRunning = true;
            }
        }
        const rotate = (percentage / 100) * 360;
        const d = describeArcAsPath(0, 0, insideRadius, 0, rotate);
        const innerRadius = insideRadius / 3;
        return svg `
        <svg style=${styleMap({
            width: "1px",
            height: "1px",
            overflow: "visible",
        })}>
            <g class="${groupClasses.join(' ')}">
                <circle cx="0" cy="0" r="${radius}" class="halo" stroke-width="${nodeStrokeWidth}" />
                <circle cx="0" cy="0" r="${insideRadius}" class="statusColor" stroke-width="${nodeStrokeWidth}" />
                <circle cx="0" cy="0" r="${innerRadius}" class="inner statusColor" />
                ${percentage ? svg `<path class="${result}" fill="none" stroke-width="${nodeStrokeWidth}" d="${d}"></path>` : svg ``}
            </g>
        </svg>
        `;
    }
};
SvgSpinner = __decorate([
    customElement(`svg-spinner`)
], SvgSpinner);
export { SvgSpinner };
//# sourceMappingURL=svg-spinner.js.map