import { LitElement, html, } from 'lit-element';
import { styleMap, } from 'lit-html/directives/style-map.js';
//--------------------------------------
//  Safety constants
//--------------------------------------
const MINLENGTH = 5; // Minimum size of cut-down text
const MAXLOOPS = 50; // Max no of iterations attempting to find the correct size text
//--------------------------------------
//  Render lifecycle
//--------------------------------------
var RenderState;
(function (RenderState) {
    RenderState[RenderState["INITIAL"] = 0] = "INITIAL";
    RenderState[RenderState["MEASURE"] = 1] = "MEASURE";
    RenderState[RenderState["FLUID"] = 2] = "FLUID";
    RenderState[RenderState["STABLE"] = 3] = "STABLE";
})(RenderState || (RenderState = {}));
/**
 * Multi-line label that will truncate with ellipses
 *
 * Use with a set width + height (or maxWidth / maxHeight) to get any use from it :D
 */
export class TruncatingLabel extends LitElement {
    constructor() {
        super(...arguments);
        this.propChildren = ``;
        this.propStyle = {};
        this.className = ``;
        //--------------------------------------
        //  Component state / lifecycle
        //--------------------------------------
        this.completeText = ''; // Unabridged plain text content
        this.innerText = ''; // Current innerText of element - includes possible ellipses
        this.renderState = RenderState.INITIAL; // Internal rendering lifecycle state
        //--------------------------------------
        //  Binary search state
        //--------------------------------------
        this.textCutoffLength = 0; // Last count used to truncate completeText
        this.longestGood = 0; // Length of the longest truncated text that fits
        this.shortestBad = 0; // Length of the shortest truncated text that does not fit
        this.loopCount = 0; // to avoid infinite iteration
    }
    // Properties
    static get properties() {
        return {
            children: { type: String, },
            style: { type: Object, },
            className: { type: String, },
        };
    }
    //--------------------------------------
    //  Lifecycle
    //--------------------------------------
    createRenderRoot() {
        return this;
    }
    connectedCallback() {
        this.handleProps({
            children: this.propChildren,
            style: this.propStyle,
            className: this.className,
        });
        super.connectedCallback();
    }
    firstUpdated() {
        this.invalidateSize();
    }
    updated() {
        this.invalidateSize();
    }
    disconnectedCallback() {
        if (this.checkSizeRequest) {
            cancelAnimationFrame(this.checkSizeRequest);
            this.checkSizeRequest = 0;
        }
        super.disconnectedCallback();
    }
    //--------------------------------------
    //  Render
    //--------------------------------------
    render() {
        const { className = '' } = this;
        const style = this.propStyle || {};
        const mergedStyle = {
            ...style,
            overflow: 'hidden',
            wordWrap: 'break-word',
        };
        if (this.renderState !== RenderState.STABLE) {
            mergedStyle.opacity = `0`;
        }
        return html `
            <div style=${styleMap(mergedStyle)} class="${'TruncatingLabel ' + className}" title="${this.innerText}">
                ${this.innerText}
            </div>
        `;
    }
    //--------------------------------------
    //  Internal Rendering Lifecycle
    //--------------------------------------
    handleProps(props) {
        const { children = '' } = props;
        if (typeof children === 'string') {
            this.completeText = children;
        }
        else if (children === null || children === false) {
            this.completeText = '';
        }
        else {
            console.warn('TruncatingLabel - Label children must be string but is', typeof children, children);
            this.completeText = 'Contents must be string';
        }
        this.renderState = RenderState.MEASURE;
        this.innerText = this.completeText;
        this.loopCount = 0;
        this.longestGood = MINLENGTH;
        this.shortestBad = this.innerText.length;
    }
    invalidateSize() {
        if (!this.checkSizeRequest) {
            this.checkSizeRequest = requestAnimationFrame(() => this.checkSize());
        }
    }
    checkSize() {
        this.checkSizeRequest = 0;
        if (this.renderState === RenderState.STABLE) {
            return; // Nothing to check, no more checks to schedule
        }
        const thisElement = this;
        const { scrollHeight, clientHeight, scrollWidth, clientWidth } = thisElement;
        const tooBig = scrollHeight > clientHeight || scrollWidth > clientWidth;
        if (this.renderState === RenderState.MEASURE) {
            // First measurement since mount / props changed
            if (tooBig) {
                this.renderState = RenderState.FLUID;
                // Set initial params for binary search of length
                this.longestGood = MINLENGTH;
                this.textCutoffLength = this.shortestBad = this.innerText.length;
            }
            else {
                this.renderState = RenderState.STABLE;
                this.requestUpdate();
            }
        }
        if (this.renderState === RenderState.FLUID) {
            this.loopCount++;
            const lastLength = this.textCutoffLength;
            let keepMeasuring;
            if (this.loopCount >= MAXLOOPS) {
                // This really shouldn't happen!
                console.error('TruncatingLabel - TOO MANY LOOPS');
                keepMeasuring = false;
            }
            else if (lastLength <= MINLENGTH) {
                keepMeasuring = false;
            }
            else if (Math.abs(this.shortestBad - this.longestGood) < 2) {
                // We're done searching, hoorays!
                keepMeasuring = false;
            }
            else {
                // Update search space
                if (tooBig) {
                    this.shortestBad = Math.min(this.shortestBad, lastLength);
                }
                else {
                    this.longestGood = Math.max(this.longestGood, lastLength);
                }
                // Calculate the next length and update the text
                this.textCutoffLength = Math.floor((this.longestGood + this.shortestBad) / 2);
                this.innerText = this.completeText.substr(0, this.textCutoffLength) + '…';
                // Bypass react's render loop during the "fluid" state for performance
                thisElement.innerText = this.innerText;
                keepMeasuring = true;
            }
            if (keepMeasuring) {
                this.invalidateSize();
            }
            else {
                this.renderState = RenderState.STABLE;
                this.requestUpdate();
            }
        }
    }
}
//# sourceMappingURL=truncating-label.js.map